<div>
  <img
    v-if="data.{{{widget.widget_id}}}.url != ''"
    :style="{
        width: data.{{{widget.widget_id}}}.width,
        height: data.{{{widget.widget_id}}}.height
    }"
    :src="data.{{{widget.widget_id}}}.url"
  />
</div>
